// Copyright (c) 2011-2012 Public Software Group e. V., Berlin, Germany
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to
// deal in the Software without restriction, including without limitation the
// rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
// sell copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
// IN THE SOFTWARE.

/*

hdrcompose is a simple command line utility to create HDR images.

Version: 0.6

Dependencies:
  ImageMagick  available at http://www.imagemagick.org/
  libtiff      available at http://www.libtiff.org/

Compile with:
gcc -O3 -ffast-math -march=native -I/usr/local/include -L/usr/local/lib -lm -ltiff -o hdrcompose hdrcompose.c `MagickWand-config --cflags --cppflags --ldflags --libs`

Replace /usr/local/include and /usr/local/lib as neccessary for libtiff.

*/


#include <stdlib.h>
#include <stdint.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

#include <wand/MagickWand.h>
#include <tiffio.h>

#define ADAPTION 0.2
#define CLIPPING 40000

#define limiter(x) ( \
    ((x) > 1.0) ? 0.75 : \
    ((x) > 0.5) ? (-(x)*(x)+2*(x)-0.25) : (x) \
  )

#define scaled_limiter(x, f) ( limiter((x) * (f)) / (f) )

#define compression(x, s) ( \
    ((s) > 0.001) ? ( \
      1.0 - pow((s) * (x) + 1.0, -1.0/(s)) \
    ) : ( \
      1.0 - exp(-(x)) \
    ) \
  )

#define GAMMA_NONE 0
#define GAMMA_SRGB 1
#define GAMMA_ADOBE 2

#define TM_NONE 0
#define TM_SRGB 1
#define TM_ADOBE 2

static int cmpfloat(float *a, float *b) {
  if (*a < *b) return -1;
  else if (*a > *b) return 1;
  else return 0;
}

int main(int argc, char **argv) {
  int input_file_count = 0;
  int width, height;
  char **filenames;
  double *ev_factors;
  float *img;
  MagickWand *input_image;
  MagickBooleanType magick_status;
  float *input_row_buffer;
  int i, j;
  size_t x, y;
  int remove_gamma, apply_gamma;
  int tm_src = TM_NONE, tm_dst = TM_NONE;
  double l = 0.0, s = 0.0, c = 0.0, k = 0.0;
  float *sorted_values;
  float minval, maxval;
  TIFF* tif;
  if (argc < 7 || (argc+1) % 2) {
    char *command;
    if (argc >= 1) command = argv[0];
    else command = "executable";
    fprintf(stderr, "Usage: %s  \\\n", command);
    fprintf(stderr, "       <input_gamma_curve> <file1> <ev1> [<file2> <ev2> ...]  \\\n");
    fprintf(stderr, "       {hdr | tm=<colorspace>:<l>:<s>:<c>:<k>}                \\\n");
    fprintf(stderr, "       <output_gamma_curve> <output_tiff_file>\n");
    fprintf(stderr, "\n");
    fprintf(stderr, "Supported gamma curves:      linear, srgb, adobe\n");
    fprintf(stderr, "Supported colorspaces:       srgb, adobe\n");
    fprintf(stderr, "                             (srgb-adobe, adobe-srgb for conversion)\n");
    fprintf(stderr, "Lightness l:                -8.0 .. +8.0   (recommended 0.0)\n");
    fprintf(stderr, "Softness s:                  0.0 .. 100.0  (recommended 1.0)\n");
    fprintf(stderr, "Chromatic adaption c:        0.0 .. 1.0    (recommended 0.0)\n");
    fprintf(stderr, "Keep color in highlights k:  0.0 .. 1.0    (recommended 0.5)\n");
    fprintf(stderr, "\n");
    fprintf(stderr, "Example: %s  \\\n", command);
    fprintf(stderr, "         srgb input-low.jpg -1 input-high.jpg 1  \\\n");
    fprintf(stderr, "         tm=srgb:0:1:0:0.5 srgb output.tiff\n", command);
    exit(1);
  }
  if      (!strcasecmp(argv[1], "linear")) remove_gamma = GAMMA_NONE;
  else if (!strcasecmp(argv[1], "srgb"  )) remove_gamma = GAMMA_SRGB;
  else if (!strcasecmp(argv[1], "adobe" )) remove_gamma = GAMMA_ADOBE;
  else {
    fprintf(stderr, "Input curve \"%s\" is not supported.\n", argv[1]);
    fprintf(stderr, "Supported curves: \"linear\", \"srgb\", \"adobe\"\n");
    exit(1);
  }
  if (strcasecmp(argv[argc-3], "hdr")) {
    char *settings;
    char *remaining;
    if (strncasecmp(argv[argc-3], "tm=", 3)) {
      fprintf(stderr, "Illegal tone mapping option \"%s\".\n", argv[argc-3]);
      exit(1);
    }
    settings = argv[argc-3] + 3;
    if (!strncasecmp(settings, "srgb", 4)) {
      tm_src = TM_SRGB;
      tm_dst = TM_SRGB;
      settings += 4;
    } else if (!strncasecmp(settings, "adobe", 5)) {
      tm_src = TM_ADOBE;
      tm_dst = TM_ADOBE;
      settings += 5;
    } else {
      fprintf(stderr, "Illegal color space for tone mapping.\n");
      exit(1);
    }
    if (*settings == '-') {
      settings++;
      if (!strncasecmp(settings, "srgb", 4)) {
        tm_dst = TM_SRGB;
        settings += 4;
      } else if (!strncasecmp(settings, "adobe", 5)) {
        tm_dst = TM_ADOBE;
        settings += 5;
      } else {
        fprintf(stderr, "Illegal color space for tone mapping.\n");
        exit(1);
      }
    }
    if (*settings != ':') {
      fprintf(stderr, "Expected colon after tonemapping color space information.\n");
      exit(1);
    }
    settings++;
    l = strtod(settings, &remaining);
    if (remaining == settings || l < -8.0 || l > 8.0) {
      fprintf(stderr, "Illegal value for l.\n");
      exit(1);
    }
    if (*remaining != ':') {
      fprintf(stderr, "Expected colon after l.\n");
      exit(1);
    }
    settings = remaining + 1;
    s = strtod(settings, &remaining);
    if (remaining == settings || s < 0.0 || s > 100.0) {
      fprintf(stderr, "Illegal value for s.\n");
      exit(1);
    }
    if (*remaining != ':') {
      fprintf(stderr, "Expected colon after s.\n");
      exit(1);
    }
    settings = remaining + 1;
    c = strtod(settings, &remaining);
    if (remaining == settings || c < 0.0 || c > 1.0) {
      fprintf(stderr, "Illegal value for c.\n");
      exit(1);
    }
    if (*remaining != ':') {
      fprintf(stderr, "Expected colon after c.\n");
      exit(1);
    }
    settings = remaining + 1;
    k = strtod(settings, &remaining);
    if (remaining == settings || k < 0.0 || k > 1.0) {
      fprintf(stderr, "Illegal value for k.\n");
      exit(1);
    }
    if (*remaining) {
      fprintf(stderr, "Expected end after k.\n");
      exit(1);
    }
  }
  if      (!strcasecmp(argv[argc-2], "linear")) apply_gamma = GAMMA_NONE;
  else if (!strcasecmp(argv[argc-2], "srgb"  )) apply_gamma = GAMMA_SRGB;
  else if (!strcasecmp(argv[argc-2], "adobe" )) apply_gamma = GAMMA_ADOBE;
  else {
    fprintf(stderr, "Output curve \"%s\" is not supported.\n", argv[argc-2]);
    fprintf(stderr, "Supported curves: \"linear\", \"srgb\", \"adobe\"\n");
    exit(1);
  }
  input_file_count = (argc-5) / 2;
  filenames = malloc(input_file_count * sizeof(char *));
  if (!filenames) {
    fprintf(stderr, "Failed to allocate memory.\n");
    exit(1);
  }
  ev_factors = malloc(input_file_count * sizeof(double));
  if (!ev_factors) {
    fprintf(stderr, "Failed to allocate memory.\n");
    exit(1);
  }
  for (i=0; i<input_file_count; i++) {
    char *ev_string;
    double ev_value;
    char *endptr;
    ev_string = argv[2+2*i+0];
    filenames[i] = ev_string;
    ev_value = strtod(argv[2+2*i+1], &endptr);
    if (*endptr) {
      fprintf(stderr, "Invalid floating point number: \"%s\"\n", ev_string);
      exit(1);
    }
    ev_factors[i] = pow(2.0, ev_value);
  }
  for (i=0; i<input_file_count-1; i++) {
    for (j=i+1; j<input_file_count; j++) {
      char *tmp1;
      double tmp2;
      if (ev_factors[i] > ev_factors[j]) {
        tmp1=filenames[i];  filenames[i] =filenames[j];  filenames[j] =tmp1;
        tmp2=ev_factors[i]; ev_factors[i]=ev_factors[j]; ev_factors[j]=tmp2;
      }
    }
  }
  MagickWandGenesis();
  for (i=0; i<input_file_count; i++) {
    int current_width, current_height;
    input_image = NewMagickWand();
    magick_status = MagickReadImage(input_image, filenames[i]);
    if (magick_status == MagickFalse) {
      char *description;
      ExceptionType severity;
      description = MagickGetException(input_image, &severity);
      fprintf(stderr, "Could not read file \"%s\": %s\n", filenames[i], description);
      exit(1);
    }
    current_width  = MagickGetImageWidth(input_image);
    current_height = MagickGetImageHeight(input_image);
    // Avoid integer overflow: 13377^2 * 3 * sizeof(float) < 2^31-1
    if (current_width > 13377 || current_height > 13377) {
      fprintf(stderr, "Image is too big.\n");
    }
    if (i == 0) {
      width  = current_width;
      height = current_height;
      input_row_buffer = malloc(3 * width * sizeof(float));
      if (!input_row_buffer) {
        fprintf(stderr, "Failed to allocate memory.\n");
        exit(1);
      }
      img = malloc(3 * width * height * sizeof(float));
      if (!img) {
        fprintf(stderr, "Failed to allocate memory.\n");
        exit(1);
      }
      for (x=0; x<3*width*height; x++) img[x] = 0.0;
    } else {
      if (current_width != width || current_height != height) {
        fprintf(stderr, "File dimensions are not matching.\n");
        exit(1);
      }
    }
    for (y=0; y<height; y++) {
      magick_status = MagickExportImagePixels(
        input_image,
        0, y, width, 1,
        "RGB", FloatPixel, input_row_buffer
      );
      if (magick_status == MagickFalse) {
        char *description;
        ExceptionType severity;
        description = MagickGetException(input_image, &severity);
        fprintf(stderr, "Error in MagickExportImagePixels call: %s\n", description);
        exit(1);
      }
      for (x=0; x<3*width; x++) {
        double value;
        value = input_row_buffer[x];
        if (remove_gamma == GAMMA_SRGB) {
          if (value <= 0.03928) value /= 12.92;
          else value = pow((value + 0.055) / 1.055, 2.4);
        } else if (remove_gamma == GAMMA_ADOBE) {
          value = pow(value, 2.19921875);
        }
        value /= ev_factors[i];
        if (i) value = scaled_limiter(value, ev_factors[i]);
        if (i < input_file_count-1) {
          value -= scaled_limiter(value, ev_factors[i+1]);
        }
        img[3*width*y+x] += value;
      }
    }
    DestroyMagickWand(input_image);
  }
  MagickWandTerminus();
  free(input_row_buffer);
  free(ev_factors);
  free(filenames);
  if (tm_src) {
    double G[3] = {0.0, 0.0, 0.0};
    double L;
    double t[3];
    if (tm_src == TM_SRGB) {
      for (x=0; x<width*height; x++) {
        for (i=0; i<3; i++) t[i] = img[3*x+i];
        img[3*x+0] = 0.4112678 * t[0] + 0.5793289 * t[1] + 0.0094033 * t[2];
        img[3*x+1] = 0.0684214 * t[0] + 0.9302672 * t[1] + 0.0013114 * t[2];
        img[3*x+2] = 0.0212514 * t[0] + 0.1177339 * t[1] + 0.8610147 * t[2];
      }
    } else if (tm_src == TM_ADOBE) {
      for (x=0; x<width*height; x++) {
        for (i=0; i<3; i++) t[i] = img[3*x+i];
        img[3*x+0] = 0.5751311 * t[0] + 0.4150863 * t[1] + 0.0097826 * t[2];
        img[3*x+1] = 0.0957661 * t[0] + 0.9028689 * t[1] + 0.0013650 * t[2];
        img[3*x+2] = 0.0297537 * t[0] + 0.0722852 * t[1] + 0.8979611 * t[2];
      }
    } else abort();
    for (x=0; x<width*height; x++) {
      for (i=0; i<3; i++) G[i] += img[3*x+i];
    }
    for (i=0; i<3; i++) G[i] /= width*height;
    L = 0.4543690 * G[0] + 0.4735332 * G[1] + 0.0720978 * G[2];
    for (i=0; i<3; i++) G[i] = (ADAPTION) * pow(2.0, l) / (c*G[i] + (1-c)*L);
    for (x=0; x<width*height; x++) {
      for (i=0; i<3; i++) t[i] = img[3*x+i] * G[i];
      img[3*x+0] =   0.7409791 * t[0] + 0.2180252 * t[1] + 0.0410057 * t[2];
      img[3*x+1] =   0.2853533 * t[0] + 0.6242016 * t[1] + 0.0904451 * t[2];
      img[3*x+2] = - 0.0096276 * t[0] - 0.0056980 * t[1] + 1.0153256 * t[2];
    }
    for (x=0; x<width*height; x++) {
      L = 0.37095 * img[3*x+0] + 0.62905 * img[3*x+1];
      L = compression(L, s) / L;
      for (i=0; i<3; i++) {
        t[i] = pow(compression(img[3*x+i], s), 1-k) * pow(L * img[3*x+i], k);
      }
      if (tm_dst == TM_SRGB) {
        img[3*x+0] =   5.435055 * t[0] - 4.598590 * t[1] + 0.163480 * t[2]; 
        img[3*x+1] = - 1.167676 * t[0] + 2.327517 * t[1] - 0.159829 * t[2];
        img[3*x+2] =   0.038035 * t[0] - 0.198699 * t[1] + 1.160664 * t[2];
      } else if (tm_dst == TM_ADOBE) {
        img[3*x+0] =   3.554233 * t[0] - 2.625705 * t[1] + 0.071436 * t[2];
        img[3*x+1] = - 1.168149 * t[0] + 2.328015 * t[1] - 0.159854 * t[2];
        img[3*x+2] = - 0.011733 * t[0] - 0.094590 * t[1] + 1.106323 * t[2];
      } else abort();
    }
  }
  sorted_values = malloc(width * height * sizeof(float));
  if (!sorted_values) {
    fprintf(stderr, "Failed to allocate memory.\n");
    exit(1);
  }
  for (x=0; x<width*height; x++) {
    sorted_values[x] = (
      (img[3*x+0] < img[3*x+1] && img[3*x+0] < img[3*x+2]) ?
      img[3*x+0] : (img[3*x+1] < img[3*x+2]) ?
      img[3*x+1] : img[3*x+2]
    );
  }
  qsort(sorted_values, width * height, sizeof(float), (void *)cmpfloat);
  minval = sorted_values[(size_t)width*height*1/(CLIPPING)];
  if (!(minval > 0.0)) minval = 0.0;
  for (x=0; x<width*height; x++) {
    sorted_values[x] = (
      (img[3*x+0] > img[3*x+1] && img[3*x+0] > img[3*x+2]) ?
      img[3*x+0] : (img[3*x+1] > img[3*x+2]) ?
      img[3*x+1] : img[3*x+2]
    );
  }
  qsort(sorted_values, width * height, sizeof(float), (void *)cmpfloat);
  maxval = sorted_values[(size_t)width*height*((CLIPPING)-1)/(CLIPPING)];
  free(sorted_values);
  maxval -= minval;
  for (x=0; x<3*width*height; x++) {
    img[x] = (img[x]-minval) / maxval;
  }
  if (apply_gamma == GAMMA_SRGB) {
    for (x=0; x<3*width*height; x++) {
      if (img[x] <= 0.00304) img[x] *= 12.92;
      else img[x] = 1.055 * pow(img[x], 1.0/2.4) - 0.055;
    }
  } else if (apply_gamma == GAMMA_ADOBE) {
    for (x=0; x<3*width*height; x++) {
      img[x] = pow(img[x], 1.0/2.19921875);
    }
  }
  tif = TIFFOpen(argv[argc-1], "w");
  if (!tif) exit(1);
  if (!TIFFSetField(tif, TIFFTAG_IMAGEWIDTH, (uint32_t)width)      ) exit(1);
  if (!TIFFSetField(tif, TIFFTAG_IMAGELENGTH, (uint32_t)height)    ) exit(1);
  if (!TIFFSetField(tif, TIFFTAG_ROWSPERSTRIP, (uint32_t)height)   ) exit(1);
  if (!TIFFSetField(tif, TIFFTAG_BITSPERSAMPLE,
    (uint16_t)8*(uint16_t)sizeof(float))                           ) exit(1);
  if (!TIFFSetField(tif, TIFFTAG_COMPRESSION, COMPRESSION_DEFLATE) ) exit(1);
  if (!TIFFSetField(tif, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_RGB)     ) exit(1);
  if (!TIFFSetField(tif, TIFFTAG_SAMPLEFORMAT, SAMPLEFORMAT_IEEEFP)) exit(1);
  if (!TIFFSetField(tif, TIFFTAG_SAMPLESPERPIXEL, (uint16_t)3)     ) exit(1);
  if (!TIFFSetField(tif, TIFFTAG_PLANARCONFIG, PLANARCONFIG_CONTIG)) exit(1);
  if (TIFFWriteEncodedStrip(
    tif, 0, img,
    3 * width * height * sizeof(float)
  ) < 0) exit(1);
  TIFFClose(tif);
  free(img);
  return 0;
}
